open Ordering
open Multiset
open Infinity

module type Atomic_ordinal_diagram_parameter = sig
    module Pi    : Setoid
    module Sigma : Setoid
  end

module type Atomic_ordinal_path_ordering_parameter = sig
    module Pi    : Well_ordering
    module Sigma : Well_ordering
  end

(** ******************************************************************
   The representation of well-ordering proofs makes somewhat messy
   code. It is only required if you want to inspect the proofs (e.g.,
   for pretty printing). Otherwise, use the rest of the interface. 
 *********************************************************************)

type ('n, 'l) atomic_ordinal_diagram 
  = Leaf of 'l
  | Node of 'n*((('n,'l) atomic_ordinal_diagram) multiset)


type 'l nested_ordinal_diagram = ('l nested_ordinal_diagram, 'l) atomic_ordinal_diagram


type ('pi, 'pi_lt, 'sigma) stratification_subterm_lt_aux
  =  {super : ('pi, 'sigma) atomic_ordinal_diagram;
      leq   : ('pi,
               ('pi, 'sigma) atomic_ordinal_diagram,
               ('pi, 'pi_lt, 'sigma) 
                 stratification_subterm_lt_aux) 
                leq
     }

type ('pi, 'pi_lt, 'sigma) stratification_subterm_lt 
  = Stratified_subterm of 
      'pi*('pi, 'pi_lt, 'sigma) stratification_subterm_lt_aux


type ('pi, 'pi_lt, 'sigma, 'sigma_lt) ordinal_path_ordering_proof_aux
  = Leaf_comparison of 
      'sigma_lt
  | Leaf_bottom  of 
      {leaf    : 'sigma
      ;nonleaf : ('pi, 'sigma) atomic_ordinal_diagram
      }
  | Subterm_extension of 
      {subterm       : ('pi, 'sigma) atomic_ordinal_diagram (* redundant, but handy *)
      ;subterm_proof : ('pi, 'pi_lt, 'sigma) stratification_subterm_lt
      ;bound_proof   : ('pi
                       ,('pi, 'sigma) atomic_ordinal_diagram
                       ,('pi, 'pi_lt, 'sigma, 'sigma_lt) ordinal_path_ordering_proof_aux) leq
      ;lesser        : ('pi, 'sigma) atomic_ordinal_diagram 
      ;greater       : ('pi, 'sigma) atomic_ordinal_diagram 
      }
  | Raise_index of
      {raised_index : 'pi infinity
      ;subterm_bounds : (('pi, 'pi_lt, 'sigma) 
                           stratification_subterm_lt
                         * ('pi, 'pi_lt, 'sigma, 'sigma_lt) 
                             ordinal_path_ordering_proof_aux) 
                          list
      ;raised_proof : ('pi, 'pi_lt, 'sigma, 'sigma_lt) 
                        ordinal_path_ordering_proof_aux
      ;lesser    : ('pi, 'sigma) atomic_ordinal_diagram
      ;greater   : ('pi, 'sigma) atomic_ordinal_diagram
      }
  | Symbol_comparison of 
      {symbol_lt : 'pi_lt
      ;lesser    : ('pi, 'sigma) atomic_ordinal_diagram
      ;greater   : ('pi, 'sigma) atomic_ordinal_diagram
      }
  | Multiset_comparison of 
      {symbol      : 'pi
      ;multiset_lt : (('pi, 'sigma) atomic_ordinal_diagram
                     ,('pi infinity * 
                         ('pi, 'pi_lt, 'sigma, 'sigma_lt) 
                           ordinal_path_ordering_proof_aux) 
                     ,('pi infinity)) parameterised_multiset_lt
      ;lesser      : ('pi, 'sigma) atomic_ordinal_diagram
      ;greater     : ('pi, 'sigma) atomic_ordinal_diagram
      }
 
type ('pi, 'pi_lt, 'sigma, 'sigma_lt) ordinal_path_ordering_proof = 
    'pi infinity * ('pi, 'pi_lt, 'sigma, 'sigma_lt) ordinal_path_ordering_proof_aux

type ('sigma, 'sigma_lt) nested_ordinal_path_ordering_proof =
  ('sigma nested_ordinal_diagram, 
   ('sigma, 'sigma_lt) nested_ordinal_path_ordering_proof,
   'sigma, 
   'sigma_lt) ordinal_path_ordering_proof
